"""Implementation of Hessians calculations.

Peter Schubert, HHU Duesseldorf, May 2022
"""

import numpy as np


def hess_from_grad_grad(grad_l, grad_r):
    """Calculate Hessian from two gradiends wrt x

    :param grad_l: left scalar gradient wrt x variables
    :param grad_l: numpy.ndarray (1D) with length of x
    :param grad_r: rigth scalar gradient wrt x variables
    :param grad_r: numpy.ndarray (1D) with length of x
    :returns: result_hess Hessian wrt to x variables, lower triangle matrix
    :rtype: numpy.ndarray (2D) with shape [x,x]
    """
    assert grad_l.shape[0] == grad_r.shape[0]
    size = grad_l.shape[0]
    result_hess = np.zeros((size, size))
    for i in range(size):
        for j in range(i+1):
            result_hess[i, j] = grad_l[i] * grad_r[j]
    return result_hess


def hess_from_v_hesss(v, hesss):
    """Calculate Hessian from a vector and Hessians

    :param v: vector of length n
    :param v: numpy.ndarray (1D)
    :param hesss: Hessians of n vectors wrt to x variables
    :param hesss: numpy.ndarray (3D) with shape [n,x,x]
    :returns: result_hess Hessian wrt to x variables, lower triangle matrix
    :rtype: numpy.ndarray (2D) with shape [x,x]
    """
    assert hesss.shape[1] == hesss.shape[2]
    size = hesss.shape[1]
    result_hess = np.zeros((size, size))
    for i in range(size):
        for j in range(i + 1):
            result_hess[i, j] = v.dot(hesss[:, i, j])
    return result_hess


def hess_from_jac_jac(jac_l, jac_r):
    """Calculate Hessian from two Jacobians

    :param jac_l: left Jacobian for vector of size n wrt to x variables
    :param jac_l: numpy.ndarray (2D) with shape [n,x]
    :param jac_r: right Jacobian for vector of size n wrt to x variables
    :param jac_r: numpy.ndarray (2D) with shape [n,x]
    :returns: result_hess Hessian wrt to x variables, lower triangle matrix
    :rtype: numpy.ndarray (2D) with shape [x,x]
    """
    assert jac_l.shape == jac_r.shape
    size = jac_l.shape[1]
    result_hess = np.zeros((size, size))
    for i in range(size):
        for j in range(i + 1):
            result_hess[i, j] = jac_l[:, i].dot(jac_r[:, j])
    return result_hess


def hesss_from_matrix_hesss(matrix, hesss):
    """Calculate dot product of a stoichiometric matrix with Hessians

    :param matrix: stoichiometric sub-matrix with shape (m, n)
    :param matrix: numpy.ndarray (2D) with shape [m,n]
    :param hesss: Hessians of n vectors wrt to x variables
    :param hesss: numpy.ndarray (3D) with shape [n,x,x]
    :returns: result_hess Hessians wrt to x variables, lower triangle matrix
    :rtype: numpy.ndarray (3D) with shape [n,x,x]
    """
    assert matrix.shape[1] == hesss.shape[0], "expect as many Hessians as there are matrix columns"
    assert hesss.shape[1] == hesss.shape[2]
    size = hesss.shape[1]
    n_rows, n_cols = matrix.shape
    result_hesss = np.zeros((n_rows, size, size))
    for i in range(n_rows):
        for j in range(n_cols):
            if matrix[i, j] != 0:
                result_hesss[i] += matrix[i, j] * hesss[j]
    return result_hesss


def hesss_from_v_hess(v, hess):
    """Calculate Hessians from vector and a Hessian

    :param v: vector of length m
    :param v: numpy.ndarray (1D)
    :param hess: Hessian of scalar wrt to x variables
    :param hess: numpy.ndarray (2D) with shape [x,x]
    :returns: result_hesss Hessians wrt to x variables, lower triangle matrix
    :rtype: numpy.ndarray (3D) with shape [n,x,x]
    """
    assert hess.shape[0] == hess.shape[1]
    size = hess.shape[1]
    n_rows = v.shape[0]

    result_hesss = np.zeros((n_rows, size, size))
    for i in range(n_rows):
        result_hesss[i] = v[i] * hess
    return result_hesss


def hesss_from_grad_jac(grad, jac):
    """Calculate Hessians from gradient and a Jacobian

    :param grad: gradient of scalar wrt to x variables
    :param grad: numpy.ndarray (1D) with shape [x]
    :param jac: Jacobian of a vector size m wrt to x variables
    :param jac: numpy.ndarray (2D) with shape [m,x]
    :returns: result_hesss Hessians wrt to x variables, lower triangle matrix
    :rtype: numpy.ndarray (3D) with shape [m,x,x]
    """
    assert grad.shape[0] == jac.shape[1]
    size = jac.shape[1]
    n_rows = jac.shape[0]

    result_hesss = np.zeros((n_rows, size, size))
    for i in range(n_rows):
        result_hesss[i] = hess_from_grad_grad(jac[i], grad)
    return result_hesss
