"""Implementation of OptResults class.

Peter Schubert, HHU Duesseldorf, June 2022
"""

import numpy as np


class OptResults:

    def __init__(self, problem, n_points):
        """Initialize OptResults.

        :param problem: GbaProblem from where to extract relevant parameters
        :type problem: GbaProblem
        :param n_points: number of points to record
        :type n_points: integer
        """
        self.problem = problem
        self._n_points = n_points
        self._idx = 0
        self._n_m = len(self.problem.var_metab_sids)
        self._n_e = len(self.problem.var_enz_sids)

        self.ress = {}
        self.ress['metabolites'] = self.problem.var_metab_sids
        self.ress['enzymes'] = self.problem.var_enz_sids
        self.ress['mr_reactions'] = self.problem.mr_rids
        self.ress['enz_sid2idx'] = {sid: idx for idx, sid in enumerate(self.problem.var_enz_sids)}
        self.ress['var_mask_metab'] = self.problem.var_mask_metab
        self.ress['var_mask_enz'] = self.problem.var_mask_enz

        self.ress['grs_per_h'] = np.zeros(n_points)
        self.ress['status'] = np.zeros(n_points)
        self.ress['x_opt'] = np.zeros((n_points, self.problem.n_vars))
        self.ress['cm_M'] = np.zeros((n_points, self._n_m))
        self.ress['rho_m'] = np.zeros((n_points, self._n_m))
        self.ress['ce_M'] = np.zeros((n_points, self._n_e))
        self.ress['rho_e'] = np.zeros((n_points, self._n_e))
        self.ress['alphas'] = np.zeros((n_points, self._n_e))
        self.ress['mr_fluxes'] = np.zeros((n_points, len(self.problem.mr_rids)))
        self.ress['psm_fluxes'] = np.zeros((n_points, self._n_e))

    def add(self, info):
        if self._idx < self._n_points:
            i = self._idx
            cx = info['x']
            self.ress['x_opt'][i] = cx
            self.ress['status'][i] = info['status']
            self.ress['grs_per_h'][i] = -3600 * info['obj_val']
            self.ress['cm_M'][i] = cx[self.problem.var_mask_metab]
            self.ress['ce_M'][i] = cx[self.problem.var_mask_enz]
            rho = self.problem.model_params['mws'] * cx
            self.ress['rho_m'][i] = rho[self.problem.var_mask_metab]
            self.ress['rho_e'][i] = rho[self.problem.var_mask_enz]

            # !!! what volume to consider, possibly the volume of the ribosome ??
            scale_factor = 3600 * 1e3 / self.problem.model_params['rho']
            self.ress['mr_fluxes'][i] = scale_factor * self.problem.mras(cx) / self.problem.var_vols[-1]
            self.ress['psm_fluxes'][i] = scale_factor / (self.problem.pstm_scale * self.problem.pstmas(cx))

            self.ress['alphas'][i] = self.problem.get_alphas(cx)
            self._idx += 1

    def get(self):
        return self.ress
